import os
from GANDLF.compute import create_pytorch_objects
from GANDLF.config_manager import ConfigManager
from GANDLF.utils import version_check, load_model, optimize_and_save_model


def post_training_model_optimization(model_path: str, config_path: str) -> bool:
    """
    CLI function to optimize a model for deployment.

    Args:
        model_path (str): Path to the model file.
        config_path (str): Path to the config file.

    Returns:
        bool: True if successful, False otherwise.
    """
    # Load the model and its parameters from the given paths
    main_dict = load_model(model_path, "cpu")
    parameters = main_dict.get("parameters", None)

    # If parameters are not available in the model file, parse them from the config file
    parameters = (
        ConfigManager(config_path, version_check_flag=False)
        if parameters is None
        else parameters
    )

    # Create PyTorch objects and set onnx_export to True for optimization
    model, _, _, _, _, parameters = create_pytorch_objects(parameters, device="cpu")
    parameters["model"]["onnx_export"] = True

    # Perform version check and load the model's state dictionary
    version_check(parameters["version"], version_to_check=main_dict["version"])
    model.load_state_dict(main_dict["model_state_dict"])

    # Optimize the model and save it to an ONNX file
    optimize_and_save_model(model, parameters, model_path, onnx_export=True)

    # Check if the optimized model file exists
    optimized_model_path = model_path.replace("pth.tar", "onnx")
    if not os.path.exists(optimized_model_path):
        print("Error while optimizing the model.")
        return False

    return True



from monai.optimizers import WarmupCosineSchedule as WCS

def warmupcosineschedule(parameters):
    parameters["scheduler"]["warmup_steps"] = parameters["scheduler"].get("warmup_steps",0.1*parameters["num_epochs"])

    return WCS(
        parameters["optimizer_object"],
        t_total = parameters["num_epochs"],
        warmup_steps = parameters["scheduler"]["warmup_steps"]
    )



from typing import Tuple, Union
import numpy as np
import torch
from skimage.color import rgb2hed, hed2rgb
from torchio.transforms.augmentation import RandomTransform
from torchio.transforms import IntensityTransform
from torchio import Subject


def hed_transform(parameters):
    return RandomHEDTransform(
        haematoxylin_sigma_range=parameters["haematoxylin_sigma_range"],
        haematoxylin_bias_range=parameters["haematoxylin_bias_range"],
        eosin_sigma_range=parameters["eosin_sigma_range"],
        eosin_bias_range=parameters["eosin_bias_range"],
        dab_sigma_range=parameters["dab_sigma_range"],
        dab_bias_range=parameters["dab_bias_range"],
        cutoff_range=parameters["cutoff_range"],
    )


class AugmenterBase:
    """Base class for patch augmentation with a hed transform"""

    def __init__(self, keyword):
        """
        Args:
            keyword (str): Short name for the transformation.
        """
        self._keyword = keyword

    ## commented the following lines because the user is never given access to these
    # @property
    # def keyword(self):
    #     """Get the keyword for the augmenter."""
    #     return self._keyword

    # def shapes(self, target_shapes):
    #     """Calculate the required shape of the input to achieve the target output shape."""
    #     return target_shapes

    # def transform(self, patch):
    #     """Transform the given patch."""
    #     return patch

    # def randomize(self):
    #     """Randomize the parameters of the augmenter."""
    #     return


class ColorAugmenterBase(AugmenterBase):
    """Base class for color patch augmentation."""

    def __init__(self, keyword):
        """
        Initialize the object.
        Args:
            keyword (str): Short name for the transformation.
        """

        # Initialize the base class.
        super().__init__(keyword=keyword)


class HedColorAugmenter(ColorAugmenterBase):
    """Apply color correction in HED color space on the RGB patch."""

    def __init__(
        self,
        haematoxylin_sigma_range: Union[tuple, None],
        haematoxylin_bias_range: Union[tuple, None],
        eosin_sigma_range: Union[tuple, None],
        eosin_bias_range: Union[tuple, None],
        dab_sigma_range: Union[tuple, None],
        dab_bias_range: Union[tuple, None],
        cutoff_range: Union[tuple, None],
    ) -> ColorAugmenterBase:
        """
        The following code is derived and inspired from the following sources: https://github.com/sebastianffx/stainlib.

        Args:
            haematoxylin_sigma_range (Union[tuple, None]): Adjustment range for the Haematoxylin channel from the [-1.0, 1.0] range where 0.0 means no change. For example (-0.1, 0.1).
            haematoxylin_bias_range (Union[tuple, None]): Bias range for the Haematoxylin channel from the [-1.0, 1.0] range where 0.0 means no change. For example (-0.2, 0.2).
            eosin_sigma_range (Union[tuple, None]): Adjustment range for the Eosin channel from the [-1.0, 1.0] range where 0.0 means no change.
            eosin_bias_range (Union[tuple, None]): Bias range for the Eosin channel from the [-1.0, 1.0] range where 0.0 means no change.
            dab_sigma_range (Union[tuple, None]): Adjustment range for the DAB channel from the [-1.0, 1.0] range where 0.0 means no change.
            dab_bias_range (Union[tuple, None]): Bias range for the DAB channel from the [-1.0, 1.0] range where 0.0 means no change.
            cutoff_range (Union[tuple, None]): Patches with mean value outside the cutoff interval will not be augmented. Values from the [0.0, 1.0] range. The RGB channel values are from the same range.

        Returns:
            ColorAugmenterBase: The color augmenter object.
        """

        # Initialize base class.
        super().__init__(keyword="hed_color")

        # Initialize members.
        self._sigma_ranges = None  # Configured sigma ranges for H, E, and D channels.
        self._bias_ranges = None  # Configured bias ranges for H, E, and D channels.
        self._cutoff_range = None  # Cutoff interval.
        self._sigmas = None  # Randomized sigmas for H, E, and D channels.
        self._biases = None  # Randomized biases for H, E, and D channels.

        # Save configuration.
        self._setsigmaranges(
            haematoxylin_sigma_range=haematoxylin_sigma_range,
            eosin_sigma_range=eosin_sigma_range,
            dab_sigma_range=dab_sigma_range,
        )
        self._setbiasranges(
            haematoxylin_bias_range=haematoxylin_bias_range,
            eosin_bias_range=eosin_bias_range,
            dab_bias_range=dab_bias_range,
        )
        self._setcutoffrange(cutoff_range=cutoff_range)

    def _setsigmaranges(
        self,
        haematoxylin_sigma_range: Union[tuple, None],
        eosin_sigma_range: Union[tuple, None],
        dab_sigma_range: Union[tuple, None],
    ):
        """
        Set the sigma intervals.

        Args:
            haematoxylin_sigma_range (Union[tuple, None]): Adjustment range for the Haematoxylin channel.
            eosin_sigma_range (Union[tuple, None]): Adjustment range for the Eosin channel.
            dab_sigma_range (Union[tuple, None]): Adjustment range for the DAB channel.
        """

        def check_sigma_range(name, given_range):
            assert given_range is None or (
                len(given_range) == 2
                and given_range[0] < given_range[1]
                and -1.0 <= given_range[0] <= 1.0
                and -1.0 <= given_range[1] <= 1.0
            ), f"Invalid range for {name}: {given_range}"

        check_sigma_range("Haematoxylin Sigma", haematoxylin_sigma_range)
        check_sigma_range("Eosin Sigma", eosin_sigma_range)
        check_sigma_range("Dab Sigma", dab_sigma_range)

        self._sigma_ranges = [
            haematoxylin_sigma_range,
            eosin_sigma_range,
            dab_sigma_range,
        ]
        self._sigmas = [
            haematoxylin_sigma_range[0]
            if haematoxylin_sigma_range is not None
            else 0.0,
            eosin_sigma_range[0] if eosin_sigma_range is not None else 0.0,
            dab_sigma_range[0] if dab_sigma_range is not None else 0.0,
        ]

    def _setbiasranges(
        self,
        haematoxylin_bias_range: Union[tuple, None],
        eosin_bias_range: Union[tuple, None],
        dab_bias_range: Union[tuple, None],
    ):
        """
        Set the bias intervals.

        Args:
            haematoxylin_bias_range (Union[tuple, None]): Bias range for the Haematoxylin channel.
            eosin_bias_range (Union[tuple, None]): Bias range for the Eosin channel.
            dab_bias_range (Union[tuple, None]): Bias range for the DAB channel.
        """

        def check_bias_range(name, given_range):
            assert given_range is None or (
                len(given_range) != 2
                or given_range[0] < given_range[1]
                or -1.0 <= given_range[0]
                or given_range[1] <= 1.0
            ), f"Invalid range for {name}: {given_range}"

        check_bias_range("Haematoxylin Bias", haematoxylin_bias_range)
        check_bias_range("Eosin Bias", eosin_bias_range)
        check_bias_range("Dab Bias", dab_bias_range)

        self._bias_ranges = [haematoxylin_bias_range, eosin_bias_range, dab_bias_range]
        self._biases = [
            haematoxylin_bias_range[0] if haematoxylin_bias_range is not None else 0.0,
            eosin_bias_range[0] if eosin_bias_range is not None else 0.0,
            dab_bias_range[0] if dab_bias_range is not None else 0.0,
        ]

    def _setcutoffrange(self, cutoff_range: Union[tuple, None]):
        """
        Set the cutoff value. Patches with mean value outside the cutoff interval will not be augmented.

        Args:
            cutoff_range (Union[tuple, None]): Cutoff range for mean value.
        """

        def check_cutoff_range(name, given_range):
            assert given_range is None or (
                len(given_range) != 2
                or given_range[0] < given_range[1]
                or 0 <= given_range[0]
                or given_range[1] <= 1.0
            ), f"Invalid range for {name}: {given_range}"

        check_cutoff_range("Cutoff", cutoff_range)

        self._cutoff_range = cutoff_range if cutoff_range is not None else [0.0, 1.0]

    ## commented the following lines because the user is never given access to this function
    # def randomize(self):
    #     """Randomize the parameters of the augmenter."""

    #     # Randomize sigma and bias for each channel.
    #     self._sigmas = [
    #         np.random.uniform(sigma_range[0], sigma_range[1]) if sigma_range else 1.0
    #         for sigma_range in self._sigma_ranges
    #     ]
    #     self._biases = [
    #         np.random.uniform(bias_range[0], bias_range[1]) if bias_range else 0.0
    #         for bias_range in self._bias_ranges
    #     ]

    def transform(self, patch: torch.Tensor) -> torch.Tensor:
        """
        Apply color deformation on the patch.

        Args:
            patch (torch.Tensor): The input patch to transform.

        Returns:
            torch.Tensor: The transformed patch.
        """

        current_patch = patch.numpy().astype(np.float32)
        patch_mean = (
            np.mean(current_patch) / 255.0
            if current_patch.dtype.kind != "f"
            else np.mean(current_patch)
        )

        if self._cutoff_range[0] <= patch_mean <= self._cutoff_range[1]:
            # Convert the image patch to HED color coding.
            patch_hed = rgb2hed(current_patch)

            # Augment the channels.
            for i in range(3):
                if self._sigmas[i] != 0.0:
                    patch_hed[..., i] *= 1.0 + self._sigmas[i]
                if self._biases[i] != 0.0:
                    patch_hed[..., i] += self._biases[i]

            # Convert back to RGB color coding.
            patch_transformed = hed2rgb(patch_hed)
            patch_transformed = np.clip(patch_transformed, 0.0, 1.0)

            # Convert back to integral data type if the input was also integral.
            if current_patch.dtype.kind != "f":
                patch_transformed *= 255.0
                patch_transformed = patch_transformed.astype(np.uint8)

            return patch_transformed

        # The image patch is outside the cutoff interval.
        return patch


class RandomHEDTransform(RandomTransform, IntensityTransform):
    def __init__(
        self,
        haematoxylin_sigma_range: Union[float, Tuple[float, float]] = 0.1,
        haematoxylin_bias_range: Union[float, Tuple[float, float]] = 0.1,
        eosin_sigma_range: Union[float, Tuple[float, float]] = 0.1,
        eosin_bias_range: Union[float, Tuple[float, float]] = 0.1,
        dab_sigma_range: Union[float, Tuple[float, float]] = 0.1,
        dab_bias_range: Union[float, Tuple[float, float]] = 0.1,
        cutoff_range: Union[float, Tuple[float, float]] = (0, 1),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.transform_object = HedColorAugmenter(
            haematoxylin_sigma_range=haematoxylin_sigma_range,
            haematoxylin_bias_range=haematoxylin_bias_range,
            eosin_sigma_range=eosin_sigma_range,
            eosin_bias_range=eosin_bias_range,
            dab_sigma_range=dab_sigma_range,
            dab_bias_range=dab_bias_range,
            cutoff_range=cutoff_range,
        )

    def apply_transform(self, subject: Subject) -> Subject:
        # Process only if the image is RGB
        for _, image in self.get_images_dict(subject).items():
            if image.data.shape[-1] == 1:
                if image.data.ndim == 4:
                    tensor = image.data[..., 0]
                    # put channel to last axis (needed for colorconv to work)
                    tensor = tensor.permute(2, 1, 0)

                    # Apply transform
                    transformed_tensor = self.transform_object.transform(tensor)

                    # Convert tensor back to tensor data
                    transformed_data = (
                        torch.from_numpy(transformed_tensor)
                        .permute(2, 0, 1)
                        .unsqueeze(-1)
                    )

                    # Update image data
                    image.set_data(transformed_data)

        return subject



# -*- coding: utf-8 -*-
"""
Implementation of functions to clip gradients
"""
from typing import Optional
import torch
from GANDLF.grad_clipping.adaptive_gradient_clipping import adaptive_gradient_clip_


def dispatch_clip_grad_(
    parameters: torch.Tensor,
    value: float,
    mode: Optional[str] = "norm",
    norm_type: Optional[float] = 2.0,
) -> None:
    """
    Dispatches the gradient clipping method to the corresponding function based on the mode.

    Args:
        parameters (torch.Tensor): The model parameters to be clipped.
        value (float): The clipping value/factor/norm, mode dependent.
        mode (Optional[str], optional): The mode of clipping. Defaults to "norm".
        norm_type (Optional[float], optional): The type of norm to compute. Defaults to 2.0.
    """
    assert mode in ["norm", "value", "agc"], f"Unknown clip mode ({mode})."
    if mode == "norm":
        torch.nn.utils.clip_grad_norm_(parameters, value, norm_type=norm_type)
    elif mode == "value":
        torch.nn.utils.clip_grad_value_(parameters, value)
    elif mode == "agc":
        adaptive_gradient_clip_(parameters, value, norm_type=norm_type)



from typing import Optional, Tuple
from pandas.util import hash_pandas_object
import torch
from torch.utils.data import DataLoader

from GANDLF.models import get_model
from GANDLF.schedulers import get_scheduler
from GANDLF.optimizers import get_optimizer
from GANDLF.data import (
    get_train_loader,
    get_validation_loader,
)
from GANDLF.utils import (
    populate_header_in_parameters,
    parseTrainingCSV,
    send_model_to_device,
    get_class_imbalance_weights,
)


def create_pytorch_objects(
    parameters: dict,
    train_csv: Optional[str] = None,
    val_csv: Optional[str] = None,
    device: Optional[str] = "cpu",
) -> Tuple[
    torch.nn.Module,
    torch.optim.Optimizer,
    DataLoader,
    DataLoader,
    torch.optim.lr_scheduler.LRScheduler,
    dict,
]:
    """
    This function creates the PyTorch objects needed for training and validation.

    Args:
        parameters (dict): The parameters for the model and training.
        train_csv (Optional[str], optional): The path to the training CSV file. Defaults to None.
        val_csv (Optional[str], optional): The path to the validation CSV file. Defaults to None.
        device (Optional[str], optional): The device to use for training. Defaults to "cpu".

    Returns:
        Tuple[ torch.nn.Module, torch.optim.Optimizer, DataLoader, DataLoader, torch.optim.lr_scheduler.LRScheduler, dict, ]: The model, optimizer, train loader, validation loader, scheduler, and parameters.
    """
    # initialize train and val loaders
    train_loader, val_loader = None, None
    headers_to_populate_train, headers_to_populate_val = None, None

    if train_csv is not None:
        # populate the data frames
        parameters["training_data"], headers_to_populate_train = parseTrainingCSV(
            train_csv, train=True
        )
        parameters = populate_header_in_parameters(
            parameters, headers_to_populate_train
        )

        # Calculate the weights here
        (
            parameters["penalty_weights"],
            parameters["sampling_weights"],
            parameters["class_weights"],
        ) = get_class_imbalance_weights(parameters["training_data"], parameters)

        print("Penalty weights : ", parameters["penalty_weights"])
        print("Sampling weights: ", parameters["sampling_weights"])
        print("Class weights   : ", parameters["class_weights"])

        # get the train loader
        train_loader = get_train_loader(parameters)
        parameters["training_samples_size"] = len(train_loader)
        # get the hash of the training data for reproducibility
        parameters["training_data_hash"] = hash_pandas_object(
            parameters["training_data"]
        ).sum()

    if val_csv is not None:
        parameters["validation_data"], headers_to_populate_val = parseTrainingCSV(
            val_csv, train=False
        )
        if headers_to_populate_train is None:
            parameters = populate_header_in_parameters(
                parameters, headers_to_populate_val
            )
        # get the validation loader
        val_loader = get_validation_loader(parameters)

    # get the model
    model = get_model(parameters)
    parameters["model_parameters"] = model.parameters()

    # get the optimizer
    optimizer = get_optimizer(parameters)
    parameters["optimizer_object"] = optimizer

    # send model to correct device
    (
        model,
        parameters["model"]["amp"],
        parameters["device"],
        parameters["device_id"],
    ) = send_model_to_device(
        model, amp=parameters["model"]["amp"], device=device, optimizer=optimizer
    )

    # only need to create scheduler if training
    if train_csv is not None:
        if not ("step_size" in parameters["scheduler"]):
            parameters["scheduler"]["step_size"] = (
                parameters["training_samples_size"] / parameters["learning_rate"]
            )

        scheduler = get_scheduler(parameters)

    else:
        scheduler = None

    # these keys contain generators, and are not needed beyond this point in params
    generator_keys_to_remove = ["optimizer_object", "model_parameters"]
    for key in generator_keys_to_remove:
        parameters.pop(key, None)

    return model, optimizer, train_loader, val_loader, scheduler, parameters



import os
from typing import Optional
import numpy as np
import tiffslide
from GANDLF.data.patch_miner.opm.utils import get_patch_size_in_microns, tissue_mask
from skimage.transform import resize
from torch.utils.data.dataset import Dataset


def get_tissue_mask(image: np.ndarray) -> np.ndarray:
    """
    This function is used to generate tissue masks; works for patches as well

    Args:
        img_rgb (np.ndarray): Input image.

    Returns:
        np.ndarray: The tissue mask.
    """
    try:
        resized_image = resize(image, (512, 512), anti_aliasing=True)
        mask = tissue_mask(resized_image)
        # upsample the mask to original size with nearest neighbor interpolation
        mask = resize(mask, (image.shape[0], image.shape[1]), order=0, mode="constant")
    except Exception as e:
        print("Entering fallback in histology inference loader because of: ", e)
        mask = np.ones(image.shape, dtype=np.ubyte)

    return mask


class InferTumorSegDataset(Dataset):
    def __init__(
        self,
        wsi_path,
        patch_size,
        stride_size,
        selected_level,
        mask_level,
        transform: Optional[object] = None,
    ):
        self.transform = transform
        self._wsi_path = wsi_path
        self._patch_size = patch_size
        self._patch_size = get_patch_size_in_microns(wsi_path, self._patch_size)
        self._stride_size = stride_size
        if self._stride_size is None:
            self._stride_size = (
                (np.array(self._patch_size) / 2).astype(np.uint16).tolist()
            )
        self._stride_size = get_patch_size_in_microns(wsi_path, self._stride_size)
        self._selected_level = selected_level
        self._mask_level = mask_level
        self._os_image = tiffslide.open_slide(os.path.join(self._wsi_path))
        self._points = []
        self._basic_preprocessing()

    def _basic_preprocessing(self):
        mask = None
        height, width = self._os_image.level_dimensions[self._selected_level]
        try:
            mask_xdim, mask_ydim = self._os_image.level_dimensions[self._mask_level]
            mask = get_tissue_mask(
                self._os_image.read_region(
                    (0, 0),
                    self._mask_level,
                    (mask_xdim, mask_ydim),
                    as_array=True,
                )
            )

            if self._selected_level != self._mask_level:
                mask = resize(mask, (height, width))
            mask = (mask > 0).astype(np.ubyte)
        except Exception as e:
            print("Mask could not be initialized, using entire image:", e)
        # This is buggy because currently if mask_level is not equal to selected_level,
        # then this logic straight up does not work
        # You would have to scale the patch size appropriately for this to work correctly
        # Remove all the points which are closer to the boundary of the wsi
        # by accessing the WSI level properties with
        # self._os_image.level_dimensions[self._selected_level]
        # Logic as if point + self.patch_size > wsi_dimensions
        # The move the point by the wsi_dimensions - (patch_size + self.points)
        # This is because the patch is not going to be extracted if it is
        # outside the wsi
        for i in range(
            0,
            width - (self._patch_size[0] + self._stride_size[0]),
            self._stride_size[0],
        ):
            for j in range(
                0,
                height - (self._patch_size[1] + self._stride_size[1]),
                self._stride_size[1],
            ):
                # If point goes beyond the wsi in y_dim, then move so that we can extract the patch
                coord_width, coord_height = i, j
                if i + self._patch_size[0] > width:
                    coord_width = width - self._patch_size[0]
                # If point goes beyond the wsi in x_dim, then move so that we can extract the patch
                if j + self._patch_size[1] > height:
                    coord_height = height - self._patch_size[1]
                # If there is anything in the mask patch, only then consider it
                if mask is not None:
                    if np.any(
                        mask[
                            coord_width : coord_width + self._patch_size[0],
                            coord_height : coord_height + self._patch_size[1],
                        ]
                    ):
                        self._points.append([coord_width, coord_height])
                else:
                    self._points.append([coord_width, coord_height])

        self._points = np.array(self._points)
        self._points[:, [0, 1]] = self._points[:, [1, 0]]

    def get_patch_size(self):
        return self._patch_size

    def __len__(self):
        return len(self._points)

    def __getitem__(self, idx):
        """
        This function is used to return the patch and its location.

        Args:
            idx (int): The index of the patch.

        Returns:
            (string, int, int): The patch, x and y locations.
        """
        x_loc, y_loc = self._points[idx]
        patch = self._os_image.read_region(
            (x_loc, y_loc),
            self._selected_level,
            (self._patch_size[0], self._patch_size[1]),
            as_array=True,
        )

        # this is to ensure that channels come at the beginning
        patch = patch.transpose([2, 0, 1])
        # this is to ensure that we always have a z-stack before applying any torchio transforms
        patch = np.expand_dims(patch, axis=-1)
        if self.transform is not None:
            patch = self.transform(patch)

        # remove z-stack
        patch = patch.squeeze(-1)

        return patch, (x_loc, y_loc)



import torch
from typing import List, Optional

from torchio.transforms.intensity_transform import IntensityTransform
from torchio.data.subject import Subject
from torchio.data.image import ScalarImage


class NormalizeRGB(IntensityTransform):
    """Threshold the intensities of the RGB images in the subject into a range.

    Args:
        mean: Expended means :math:`[a, b, c]` of the output image.
        std: Expected std-deviation :math:`[a', b', c']` of the output image.

    """

    def __init__(
        self,
        mean: Optional[List[float]] = None,
        std: Optional[List[float]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.mean, self.std = mean, std
        self.args_names = "mean", "std"

    def apply_transform(self, subject: Subject) -> Subject:
        for image in self.get_images(subject):
            self.apply_normalize(image)
        return subject

    def apply_normalize(self, image: ScalarImage) -> None:
        image_data = image.data
        if image_data.shape[-1] == 1:
            image_data = image_data.squeeze(-1)
            image_data = self.normalize(image_data, self.mean, self.std)
            image_data = image_data.unsqueeze(-1)
            image.set_data(image_data)
        else:
            image.set_data(self.normalize(image_data, self.mean, self.std))

    def normalize(self, tensor: torch.Tensor, mean: List[float], std: List[float]):
        """Normalize a float tensor image with mean and standard deviation.
        This transform does not support PIL image.

        Args:
            tensor (Tensor): Float tensor image of size (C, H, W) or (B, C, H, W) to be normalized.
            mean (sequence): Sequence of means for each channel.
            std (sequence): Sequence of standard deviations for each channel.

        Returns:
            Tensor: Normalized Tensor image.
        """

        # standard operation defined in ToTensor
        tensor = tensor.div(255)
        dtype = tensor.dtype
        mean = torch.as_tensor(mean, dtype=dtype, device=tensor.device)
        std = torch.as_tensor(std, dtype=dtype, device=tensor.device)

        if (std == 0).any():
            raise ValueError(
                f"std evaluated to zero after conversion to {dtype}​​​​​, leading to division by zero."
            )
        if mean.ndim == 1:
            mean = mean.view(-1, 1, 1)
        if std.ndim == 1:
            std = std.view(-1, 1, 1)

        tensor = tensor.sub(mean)
        tensor = tensor.div(std)

        return tensor


# the "_transform" functions return lambdas that can be used to wrap into a Compose class
def normalize_by_val_transform(mean, std):
    return NormalizeRGB(mean=mean, std=std)


def normalize_imagenet_transform():
    return NormalizeRGB(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])


def normalize_standardize_transform():
    return NormalizeRGB(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])


def normalize_div_by_255_transform():
    return NormalizeRGB(mean=[0.0, 0.0, 0.0], std=[1.0, 1.0, 1.0])
